"""Benchmarktest from Crombecq et al. (2011). Benchmarks LV and
Probabilistic sampling vs random sampling. """
# import matplotlib.pyplot as plt
# import numpy as np
#
# from harlow.lola_voronoi import LolaVoronoi
# from harlow.surrogate_model import GaussianProcess
# from tests.integration_tests.test_functions import shekel


def test_LV_sampling():
    pass


def test_probabilistic_sampling():
    pass


def test_random_sampling():
    pass
